SystemOrganization addCategory: #'GT-BytecodeDebugger'!
SystemOrganization addCategory: 'GT-BytecodeDebugger-Actions'!
SystemOrganization addCategory: 'GT-BytecodeDebugger-Model'!
SystemOrganization addCategory: 'GT-BytecodeDebugger-UI'!

GTDebuggerVariablesBrowser subclass: #GTBytecodeDebuggerVariablesBrowser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-BytecodeDebugger-UI'!

!GTBytecodeDebuggerVariablesBrowser commentStamp: 'AndreiChis 12/30/2015 11:03' prior: 0!
I provide a different view of all variables from  a Context suitable for a bytecode debugger:
- I show the complete stack of variables stored by the context (both named and unnamed variables)
- I show the slots of the receiver object using the notation rcv0, rcv1, etc., as this is how slots are accessed in bytecode instructions!

!GTBytecodeDebuggerVariablesBrowser methodsFor: 'as yet unclassified' stamp: 'AndreiChis 1/14/2015 19:36'!
variableTypeOf: aString in: aContext
	|variableName|
	
	variableName := (aString includes: $[)
			ifTrue: [ (aString copyFrom: (aString indexOf: $[) +1 to: (aString indexOf: $])-1) trim  ]
			ifFalse: [ aString ].
	^ super 
		variableTypeOf: variableName
		in: aContext! !

!GTBytecodeDebuggerVariablesBrowser methodsFor: 'private building' stamp: 'AndreiChis 1/14/2015 19:07'!
variablesFromContext: aContext 

	^ aContext gtBytecodeDebuggerRetrieveContextValues! !

!GTBytecodeDebuggerVariablesBrowser methodsFor: 'private building' stamp: 'AndreiChis 1/14/2015 19:02'!
variablesIn: aComposite

	^ (super variablesIn: aComposite)
		sorted: nil! !

!Context methodsFor: '*GT-BytecodeDebugger' stamp: 'AndreiChis 1/16/2015 15:51'!
gtBytecodeDebuggerRetrieveContextValues
	|fieldList displayedTemp context slotsSize|

	fieldList := OrderedCollection new.
	context := self.
	context == nil ifTrue: [^('_thisContext' -> nil) asOrderedCollection].
	fieldList
		add: ('_self' ->  context receiver);
		add: ('_thisContext' ->  context).
	
	displayedTemp := OrderedCollection new.
	context stackPtr to: 1 by: -1 do: [ :index ||key|
		key := (index-1) asString.
		(index <= context tempNames size) ifTrue: [ 
			key := key, ' [', (context tempNames at: index) asString, ']'.
			displayedTemp add: (context tempNames at: index) ].
		fieldList add: (key -> (context at: index)) ].
	(context tempNames copy asOrderedCollection removeAll: displayedTemp; yourself) ifNotEmptyDo: [ :uninitializedTemps |
		uninitializedTemps do: [ :each |
			fieldList add: ('- [', each, ']') -> nil ] ].
	slotsSize := context receiver class allSlots size.
	fieldList 
		addAll: ((context receiver class allSlots reverse) collect: [:slot ||name| 
			name := slot name.
			('rcv', (slotsSize := slotsSize - 1) asString, ' [', name , ']') -> (slot read: context receiver) ]).
	
	^ fieldList! !

DebugSession subclass: #GTBytecodeDebuggerSession
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-BytecodeDebugger-Model'!

!GTBytecodeDebuggerSession commentStamp: 'AndreiChis 12/30/2015 10:51' prior: 0!
I provide debugging actions at the level of the bytecode.
I also make sure  that step over only goes over one single bytecode instruction.!

!GTBytecodeDebuggerSession methodsFor: 'debugging actions' stamp: 'AndreiChis 1/16/2015 16:18'!
runToBytecode: aSymbolicBytecode inContext: aContext 
	"Attempt to step over instructions in selectedContext until the 
	execution reaches the selected instruction. This happens when the
	program counter passes the begining of selectionInterval." 
	
	aSymbolicBytecode offset <= aContext pc 
		ifTrue: [ ^self ].
	self stepOver: aContext.
	[ aContext == self interruptedContext and: [ aContext pc < aSymbolicBytecode offset ] ] 
		whileTrue: [ self stepOver: aContext ]! !

!GTBytecodeDebuggerSession methodsFor: 'private' stamp: 'AndreiChis 1/16/2015 16:15'!
stepToFirstInterestingBytecodeIn: aProcess
	"Do not step over any bytecode."
	
	^ aProcess suspendedContext! !

!GTBytecodeDebuggerSession methodsFor: 'acessing' stamp: 'AndreiChis 10/22/2014 14:47'!
symbolicBytecodeForCurrentPCFrom: aContext
	| pc |
	
	pc := aContext pc.
	^ aContext method symbolicBytecodes 
		detect: [ :each | each offset = pc ]
		ifNone: [ nil ]! !

DebugAction subclass: #GTGoAndInspectBytecodeDebugAction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-BytecodeDebugger-Actions'!

!GTGoAndInspectBytecodeDebugAction commentStamp: 'AndreiChis 12/30/2015 10:53' prior: 0!
I am a debugging action that inspects the currently selected bytecode.
!

!GTGoAndInspectBytecodeDebugAction class methodsFor: 'registration' stamp: 'AndreiChis 9/8/2015 17:25'!
gtBytecodeDebuggerActionFor: aDebugger
	<gtContextBytecodeCodeDebuggingAction>
	
	^ (self forDebugger: aDebugger)! !

!GTGoAndInspectBytecodeDebugAction methodsFor: 'as yet unclassified' stamp: 'AndreiChis 1/14/2015 22:12'!
defaultKeyText

	^ 'g'! !

!GTGoAndInspectBytecodeDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 10/23/2014 15:15'!
defaultLabel

	^ 'Go and inspect'! !

!GTGoAndInspectBytecodeDebugAction methodsFor: 'actions' stamp: 'AndreiChis 1/15/2015 16:29'!
executeAction
	
	self debugger inspect: self debugger bytecodePresentation selection  
		
! !

!GTGoAndInspectBytecodeDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 1/14/2015 22:08'!
id

	^ GTGoAndInspectBytecodeDebugAction name! !

!GTGoAndInspectBytecodeDebugAction methodsFor: 'initialization' stamp: 'AndreiChis 10/23/2014 15:18'!
initialize

	super initialize.
	self needsUpdate: false! !

DebugAction subclass: #GTStepToBytecodeDebugAction
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-BytecodeDebugger-Actions'!

!GTStepToBytecodeDebugAction commentStamp: 'AndreiChis 12/30/2015 10:52' prior: 0!
I am a debugging action for stepping to a selected bytecode.!

!GTStepToBytecodeDebugAction class methodsFor: 'registration' stamp: 'AndreiChis 9/8/2015 17:25'!
gtBytecodeDebuggerActionFor: aDebugger
	<gtContextBytecodeCodeDebuggingAction>
	
	^ (self forDebugger: aDebugger)
			icon: GLMUIThemeExtraIcons glamorousPlay! !

!GTStepToBytecodeDebugAction methodsFor: 'testing' stamp: 'AndreiChis 1/15/2015 16:29'!
appliesToDebugger: aDebugger

	^ aDebugger session isInterruptedContextPostMortem not and: [ 
		aDebugger bytecodePresentation selection offset > aDebugger currentContext pc ]! !

!GTStepToBytecodeDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 1/14/2015 22:46'!
defaultLabel

	^ 'Step to bytecode'! !

!GTStepToBytecodeDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 1/14/2015 22:46'!
defaultOrder

	^ 10! !

!GTStepToBytecodeDebugAction methodsFor: 'actions' stamp: 'AndreiChis 1/15/2015 16:29'!
executeAction

	self session
		runToBytecode: self debugger bytecodePresentation selection 
		inContext: self currentContext! !

!GTStepToBytecodeDebugAction methodsFor: 'accessing' stamp: 'AndreiChis 1/14/2015 22:46'!
id

	^ GTStepToBytecodeDebugAction name! !

GTGenericStackDebugger subclass: #GTBytecodeDebugger
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GT-BytecodeDebugger-UI'!

!GTBytecodeDebugger commentStamp: 'AndreiChis 12/30/2015 10:49' prior: 0!
I am a debugger that provides the ability to step through the execution at the bytecode level (e.g. one bytecode at a time).!

!GTBytecodeDebugger class methodsFor: 'accessing' stamp: 'AndreiChis 1/10/2016 17:42'!
defaultTitle

	^ 'Bytecode'! !

!GTBytecodeDebugger class methodsFor: 'tools registry' stamp: 'AndreiChis 10/22/2014 11:13'!
register

	self registerToolsOn: Smalltalk tools.! !

!GTBytecodeDebugger class methodsFor: 'tools registry' stamp: 'AndreiChis 1/20/2015 15:02'!
registerToolsOn: registry

	"Add ourselves to registry. See [Smalltalk tools]" 
	registry registerDebugger: self withRank: 5.! !

!GTBytecodeDebugger class methodsFor: 'accessing' stamp: 'AndreiChis 10/22/2014 14:17'!
sessionClass

	^ GTBytecodeDebuggerSession! !

!GTBytecodeDebugger class methodsFor: 'accessing' stamp: 'AndreiChis 1/14/2015 18:55'!
variablesBrowserClass

	^ GTBytecodeDebuggerVariablesBrowser! !

!GTBytecodeDebugger methodsFor: 'building actions' stamp: 'AndreiChis 1/20/2015 22:25'!
bytecodeActionsPragmas

	^ #( bytecodeCodeDebuggingAction )! !

!GTBytecodeDebugger methodsFor: 'building presentations' stamp: 'AndreiChis 1/16/2015 15:58'!
bytecodeIn: aComposite forContext: aContext

	(self methodBytecodeWidgetIn: aComposite forContext: aContext)
		initialize: [ :bytecodePresentation |
				bytecodePresentation selection: (self session symbolicBytecodeForCurrentPCFrom: self currentContext) ];
		with: [ :bytecode | self installActionsForRootPragmas: self bytecodeActionsPragmas in: bytecode ].! !

!GTBytecodeDebugger methodsFor: 'accessing' stamp: 'AndreiChis 1/16/2015 15:58'!
bytecodePane

	^ self browser paneNamed: #bytecode! !

!GTBytecodeDebugger methodsFor: 'accessing' stamp: 'AndreiChis 1/16/2015 15:58'!
bytecodePresentation
	
	^ self bytecodePane presentations first! !

!GTBytecodeDebugger methodsFor: 'building' stamp: 'AndreiChis 1/10/2016 17:41'!
debuggerStructureIn: browser
	
	browser
		row: #stack;
		row: [ :c | c
			column: #code; 
			column: #bytecode ] span: 2;
		row: #inspector.
		! !

!GTBytecodeDebugger methodsFor: 'building' stamp: 'AndreiChis 1/16/2015 15:58'!
debuggerTransmissionsIn: browser

	super debuggerTransmissionsIn: browser.
	browser transmit
		from: #stack port: #selection;
		to: #bytecode;
		andShow: [ :composite :aContext | self bytecodeIn: composite forContext: aContext ]
! !

!GTBytecodeDebugger methodsFor: 'building presentations' stamp: 'AndreiChis 10/14/2015 11:34'!
decompiledCodeWidgetIn: composite forContext: aContext

	^ composite pharoMethod
		title: 'Decompiled';
		format: [ aContext method codeForNoSource ];
		smalltalkClass: [  aContext methodClass ];
		doItReceiver: [ aContext receiver ];
		doItContext: [  aContext ]! !

!GTBytecodeDebugger methodsFor: 'printing/formatting' stamp: 'AndreiChis 1/12/2016 16:15'!
formatBytecode: aBytecode
	| textAttributes|
	
	textAttributes := {TextColor new color: Smalltalk ui theme textColor}.
	self selectedContext pc = aBytecode offset ifTrue: [ textAttributes := {TextColor new color: Smalltalk ui theme textColor. TextEmphasis italic} ]. 
	self selectedContext pc > aBytecode offset ifTrue: [ textAttributes := {TextColor new color: Color gray} ].
	^ Text 
		string: aBytecode printString
		attributes: textAttributes! !

!GTBytecodeDebugger methodsFor: 'building presentations' stamp: 'AndreiChis 1/16/2015 18:02'!
methodBytecodeWidgetIn: aComposite forContext: aContext

	^ aComposite tree
		title: 'Bytecode';
		rootsExpanded;
		display: [ aContext method symbolicBytecodes ];
		children: [ :each | each isArray ifTrue: [each] ifFalse: #() ];
		format: [ :each | self formatBytecode: each ];
		onChangeOfPort: #selection act: [ :bytecodes | 
			self codePresentation 
				selectionInterval: (bytecodes selection
					ifNil: [ 1 to: 0 ] 
					ifNotNil: [:selectedBytecode| selectedBytecode sourceInterval]) ] ! !

!GTBytecodeDebugger methodsFor: 'building presentations' stamp: 'AndreiChis 1/27/2015 17:19'!
methodCodeIn: composite forContext: aContext

	(self methodCodeWidgetIn: composite forContext: aContext)
		with: [ :code | self installCodeActionsFor: code ].
	"self decompiledCodeWidgetIn: composite forContext: aContext"! !

!SymbolicBytecode methodsFor: '*GT-BytecodeDebugger' stamp: 'AndreiChis 1/14/2015 17:10'!
= anObject

	self == anObject
		ifTrue: [^ true].

	(self class == anObject class)
		ifFalse: [^ false].

	^ (self bytes = anObject bytes) and: [ 
		(self description = anObject description) and: [ 
			(self method = anObject method) and: [ 
				self offset = anObject offset ] ] ]! !

!SymbolicBytecode methodsFor: '*GT-BytecodeDebugger' stamp: 'AndreiChis 1/8/2016 15:47'!
hash

	^ ((self bytes hash bitXor: self description hash) bitXor: 
			self method hash) bitXor: self offset hash ! !
